"""Filter time series data or event frames."""

from datetime import datetime
from typing import Optional

from kukur import InterpolationType

try:
    import pandas as pd

    from timeseer_client.internal import filters_pandas

    HAS_PANDAS = True
except Exception:
    HAS_PANDAS = False

from timeseer_client.internal import (
    AugmentationStrategy,
    UnknownAugmentationStrategyException,
    filters_arrow,
)


def filter_series(
    series,
    event_frames,
    augmentation_strategy: AugmentationStrategy = AugmentationStrategy.REMOVE,
    interpolation_type: Optional[InterpolationType] = None,
    context=None,
):
    """Filter the time series in the time periods given by event_frames.

    Args:
        series: a pyarrow Table or a pandas DataFrame with time series date
            Two columns are present: 'ts' and 'value'.
            A pandas DataFrame can contain a DatetimeIndex instead of the 'ts' column.
        event_frames: pyarrow Table or a pandas DataFrame with event frames.
            Three columns need to be present: 'type', 'start_date' and 'end_date'.
        augmentation_strategy: An enum to define which strategy to use when filtering.
            'REMOVE' (the default) removes the values, 'HOLD_LAST' keeps the last acceptable value,
            'LINEAR_INTERPOLATION' interpolates the last acceptable value and the next acceptable value
            and 'MEDIAN' interpolates with the median value.
            If no acceptable value exists, they are removed. 'KNN_IMPUTATION' uses context to find the
            k-nearest neighbors and takes the average.
        interpolation_type: Enum to define the interpolation type. 'LINEAR' or 'STEPPED'.
            Only linear interpolation types can be linearly interpolated in the augmentation strategy.
        context: A list of pd.DataFrame used for 'KNN_IMPUTATION' to find nearest neighbors
    Returns:
        A filtered pyarrow Table or a pandas DataFrame with 2 columns: 'ts' and 'value'.
        In case the pandas DataFrame provided in 'series' has a DatetimeIndex,
            the 'ts' column will not be present, but the DataFrame will have a DateTimeIndex.
    """
    if context is None:
        context = []

    if not isinstance(augmentation_strategy, AugmentationStrategy):
        raise UnknownAugmentationStrategyException()

    if (
        augmentation_strategy == AugmentationStrategy.LINEAR_INTERPOLATION
        and interpolation_type != InterpolationType.LINEAR
    ):
        augmentation_strategy = AugmentationStrategy.HOLD_LAST

    if len(series) == 0:
        return series

    if (
        HAS_PANDAS
        and isinstance(series, pd.DataFrame)
        and isinstance(event_frames, pd.DataFrame)
    ):
        return filters_pandas.filter_series(
            series, event_frames, augmentation_strategy, context
        )
    return filters_arrow.filter_series(series, event_frames, augmentation_strategy)


def filter_event_frames(event_frames, start_date: datetime, end_date: datetime):
    """Restrict the event frames to the given time range.

    Args:
        event_frames: a pyarrow Table or a pandas DataFrame with event frames.
        start_date: the start date of the range to filter event_frames.
        end_date: the end date of the range to filter event_frames.

    Returns::
        A filtered pyarrow Table or a pandas DataFrame with 5 columns.
        The first column ('start_date') contains the 'start_date' and 'end_date'.
        The second column ('end_date') contains the 'end_date'.
        The third column ('type') contains the type of the returned event frame as a string.
        Columns 4 ('series_source') and 5 ('series_name') contain the source and name of the series.
    """
    if HAS_PANDAS and isinstance(event_frames, pd.DataFrame):
        return filters_pandas.filter_event_frames(event_frames, start_date, end_date)
    return filters_arrow.filter_event_frames(event_frames, start_date, end_date)



"""Extra typed metadata fields."""

from typing import Any, Optional

from kukur.metadata import MetadataField
from kukur.metadata.fields import (  # noqa: F401
    Accuracy,
    AccuracyPercentage,
    DataType,
    Description,
    Dictionary,
    DictionaryName,
    InterpolationType,
    LimitHighFunctional,
    LimitHighPhysical,
    LimitLowFunctional,
    LimitLowPhysical,
    Unit,
)

from timeseer_client.internal import ProcessType as TSProcessType


def _process_type_to_json(process_type: Optional[TSProcessType]) -> Optional[str]:
    if process_type is None:
        return None
    return process_type.value


def _process_type_from_json(process_type: Optional[str]) -> Optional[TSProcessType]:
    if process_type is None:
        return None
    return TSProcessType(process_type)


def _parse_float(number: Optional[Any]) -> Optional[float]:
    if number is None or number == "":
        return None
    return float(number)


ProcessType = MetadataField[Optional[TSProcessType]](
    "process type",
    default=None,
    serialized_name="processType",
    serialize=_process_type_to_json,
    deserialize=_process_type_from_json,
)


SensorModel = MetadataField[Optional[str]](
    "sensor model", default=None, serialized_name="sensorModel"
)


SamplingRate = MetadataField[Optional[float]](
    "sampling rate",
    default=None,
    serialized_name="samplingRate",
    deserialize=_parse_float,
)


def register_custom_fields(cls):
    """Register all typed metadata fields with the given class."""
    cls.register_field(ProcessType)
    cls.register_field(SensorModel, after_field=LimitHighFunctional)
    cls.register_field(SamplingRate, after_field=AccuracyPercentage)


__all__ = [
    "Accuracy",
    "AccuracyPercentage",
    "DataType",
    "Description",
    "Dictionary",
    "DictionaryName",
    "InterpolationType",
    "LimitHighFunctional",
    "LimitLowFunctional",
    "LimitHighPhysical",
    "LimitLowPhysical",
    "ProcessType",
    "SamplingRate",
    "SensorModel",
    "Unit",
]



"""Timeseer Client provides the low-level connection to Timeseer."""

import json
from typing import Any, Dict, List, Optional, Tuple, Union

import pyarrow as pa
import pyarrow.flight as fl
from kukur import Metadata, SeriesSelector
from kukur.client import Client as KukurClient

from timeseer_client.internal import ServerReturnedException
from timeseer_client.internal.flows import (
    evaluate_flow,
    wait_for_flow_evaluation,
)


class Client(KukurClient):
    """Client connects to Timeseer using Arrow Flight."""

    def upload_data(
        self,
        metadata_or_data: Union[Metadata, List[Tuple[Metadata, pa.Table]]],
        table: Optional[pa.Table] = None,
        *,
        analyze=True,
        block=True,
    ):
        """Upload time series data to Timeseer.

        This requires a configured 'flight-upload' source in Timeseer.

        There are two ways to call this method.

        One requires two arguments:
            metadata: any known metadata about the time series. This will be merged with the metadata
                already known by Timeseer depending on the source configuration. The source of the series should match
                the source name of a 'flight-upload' source.
            table: a pyarrow.Table of two columns.
                The first column with name 'ts' contains Arrow timestamps.
                The second column with name 'value' contains the values as a number or string.

        The second accepts a list of tuples of the same arguments. This allows uploading multiple time series at the
        same time.

        When `analyze` is `True`, start a flow evaluation.
        When `block` is `True`, block execution until the flow evaluation is done.
        """
        if table is not None:
            assert isinstance(metadata_or_data, Metadata)
            self._upload_data_single(metadata_or_data, table, analyze, block)
        else:
            assert not isinstance(metadata_or_data, Metadata)
            self._upload_data_multiple(metadata_or_data, analyze, block)

    def _upload_data_single(
        self, metadata: Metadata, table: pa.Table, analyze: bool, block: bool
    ):
        self._upload_data_multiple([(metadata, table)], analyze, block)

    def _upload_data_multiple(
        self, many_series: List[Tuple[Metadata, pa.Table]], analyze: bool, block: bool
    ):
        client = self._get_client()
        selectors = []
        for metadata, table in many_series:
            metadata_json = metadata.to_data()
            selector = SeriesSelector.from_tags(
                metadata.series.source, metadata.series.tags, metadata.series.field
            )
            selectors.append(selector)
            descriptor = fl.FlightDescriptor.for_command(
                json.dumps(
                    {
                        "metadata": metadata_json,
                    }
                )
            )
            writer, reader = client.do_put(descriptor, table.schema)
            writer.write_table(table)
            writer.done_writing()
            buf: pa.Buffer = reader.read()
            response: Dict = json.loads(buf.to_pybytes())
            writer.close()

        if "flowName" in response:
            if analyze:
                evaluate_flow(
                    self, response["flowName"], limitations=selectors, block=block
                )
        elif block and "flowEvaluationGroupId" in response:
            wait_for_flow_evaluation(self, response)

    def do_action(self, name: str, data: Any) -> Any:
        """Perform an Arrow Flight action with the given data provided as JSON.

        Args:
            name: The name of the action to perform.
            data: The json-convertible data to provide to the action.

        Returns:
            An action specific return values or nothing.
        """
        response = self._get_client().do_action((name, json.dumps(data).encode()))
        if response is not None:
            results = list(response)
            if len(results) > 0:
                return json.loads(results[0].body.to_pybytes())
        return None

    def do_get(self, data: Any) -> pa.Table:
        """Do an Arrow Flight GET request to return an Arrow table.

        Args:
            data: The json-convertible data for the Flight ticket.

        Returns:
            A PyArrow Table.
        """
        ticket = fl.Ticket(json.dumps(data))
        return self._get_client().do_get(ticket).read_all()

    def do_put(self, data: Any, table: pa.Table) -> Any:
        """Do an Arrow Flight PUT request to upload an Arrow table.

        Args:
            data: The json-convertible data for the Flight descriptor.
            table: The pyarrow.Table to PUT.

        Returns:
            The json-converted result of the PUT operation.

        Raises:
            ServerReturnedException when the server returns an error in the response
            body instead of the result of the PUT operation.
        """
        descriptor = fl.FlightDescriptor.for_command(json.dumps(data))
        writer, reader = self._get_client().do_put(descriptor, table.schema)
        writer.write_table(table)
        writer.done_writing()
        buf: pa.Buffer = reader.read()
        response: Dict = json.loads(buf.to_pybytes())
        writer.close()
        if error := response.get("error"):
            raise ServerReturnedException(error)
        return response



"""Python client for Timeseer resources."""

import json
import time
from pathlib import Path
from typing import Dict, List, Optional, Union

import kukur.config

try:
    import yaml

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

from timeseer_client.internal import (
    JSONFlightClient,
    MissingModuleException,
    TimeoutException,
)


class Resources:
    """Resources are the main organizational structures within Timeseer.

    Args:
        client: the Timeseer Client
    """

    __client: JSONFlightClient

    def __init__(self, client: JSONFlightClient):
        self.__client = client

    def list(self, *, resource_type: Optional[str] = None) -> List[Dict]:
        """List all defined resources.

        Args:
            resource_type: Only list resources of this type (optional, keyword-only)

        Returns:
            A list of dictionaries of resource 'name' and 'type'.
        """
        body = {}
        if resource_type is not None:
            body["resourceType"] = resource_type
        return self.__client.do_action("resources/list", body)

    def create(
        self,
        resources: Optional[List[Dict]] = None,
        *,
        resource: Optional[Dict] = None,
        path: Optional[Union[Path, str]] = None,
    ):
        """Create or update resources by supplying one resource, multiple resources or a filename.

        Args:
            resources: A list containing the resource definitions.
            resource: One resource definition.
            path: A path to the file or filename of a resource definition.
                This can be in yaml, toml or json format.
        """
        all_resources = []
        if resources is not None:
            assert isinstance(resources, List)
            all_resources.extend(resources)
        if resource is not None:
            assert isinstance(resource, Dict)
            all_resources.append(resource)
        if path is not None:
            if isinstance(path, str):
                path = Path(path)
            if path.suffix in [".yml", ".yaml"]:
                if not HAS_YAML:
                    raise MissingModuleException("PyYAML")
                all_resources.extend(_read_yaml(path))
            elif path.suffix == ".toml":
                all_resources.extend(_read_toml(path))
            elif path.suffix == ".json":
                all_resources.extend(_read_json(path))
        body = {
            "resources": all_resources,
        }

        self.__client.do_action("resources/create", body)

    def read(self, resource: Dict[str, str]) -> Dict:
        """Return a resource definition.

        Args:
            resource: Dictionary of resource 'type' and 'name'

        Returns:
            The resource definition.
        """
        return self.__client.do_action("resources/read", resource)

    def delete(
        self,
        resources: Optional[List[Dict]] = None,
        *,
        resource: Optional[Dict] = None,
        timeout_seconds: int = 60,
    ):
        """Remove resources.

        Some resources are removed asynchronously. This can prevent removal of other
        resources that are still in use. This method retries removal for the specified time.

        Args:
            resources: A list containing the resources to remove.
            resource: A dictionary of the resource to remove, type and name have to be present.
            timeout_seconds: The timeout before failing (optional, keyword-only).
        """
        all_resources = []
        if resources is not None:
            assert isinstance(resources, List)
            all_resources.extend(resources)

        if resource is not None:
            assert isinstance(resource, Dict)
            all_resources.append(resource)

        body = {
            "resources": all_resources,
        }

        message = None
        while timeout_seconds > 0:
            response = self.__client.do_action("resources/delete", body)
            if response is None:
                return
            message = response.get("message")
            timeout_seconds = timeout_seconds - 1
            time.sleep(1)

        raise TimeoutException(message)


def _read_yaml(path: Path) -> List[Dict]:
    with path.open() as file:
        return yaml.safe_load(file)


def _read_toml(path) -> List[Dict]:
    config = kukur.config.from_toml(path)
    return config["resource"]


def _read_json(path: Path) -> List[Dict]:
    with path.open() as file:
        return json.load(file)



"""Test filtering time series in pyarrow Tables."""

import string
from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pyarrow as pa
from kukur import InterpolationType

from timeseer_client import AugmentationStrategy, filter_series

START_DATE = datetime.fromisoformat("2020-01-01T00:00:00.000+00:00")


def _create_series_data(timestamps, values, quality=None) -> pa.Table:
    if quality is not None:
        df = pd.DataFrame(data={"ts": timestamps, "value": values, "quality": quality})
    else:
        df = pd.DataFrame(data={"ts": timestamps, "value": values})
    df = df.set_index("ts")
    return pa.Table.from_pandas(df)


def _create_event_frame_data(start_dates, end_dates):
    series_type = pa.struct(
        [
            ("series_source", pa.string()),
            ("series_name", pa.string()),
        ]
    )
    references_type = pa.list_(series_type)
    types = []
    references = []
    if len(start_dates) > 0:
        types.append("type")
        references.append(None)
    return pa.Table.from_arrays(
        [
            pa.array(start_dates, type=pa.timestamp("ns", tz="+00:00")),
            pa.array(end_dates, type=pa.timestamp("ns", tz="+00:00")),
            pa.array(types),
            pa.array(references, type=references_type),
        ],
        ["start_date", "end_date", "type", "reference"],
    )


def test_filter_empty_series() -> None:
    series = _create_series_data([], [])
    event_frames = _create_event_frame_data(
        [START_DATE], [START_DATE + timedelta(days=10)]
    )
    result = filter_series(series, event_frames)
    assert len(result) == 0


def test_filter_empty_event_frames() -> None:
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], np.random.normal(size=10)
    )
    event_frames = _create_event_frame_data([], [])
    result = filter_series(series, event_frames)
    assert len(result) == 10


def test_filter_all() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE], [START_DATE + timedelta(days=10)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == 0


def test_filter_none() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=11)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == len(series)


def test_filter_one() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == 1
    assert result["ts"][0].as_py() == START_DATE


def test_filter_one_with_quality() -> None:
    values = np.random.normal(size=10)
    quality = [0, 1] * 5
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == 1
    assert result["ts"][0].as_py() == START_DATE
    assert result["quality"][0].as_py() == series["quality"][0].as_py()


def test_hold_last_value_filter_all() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE], [START_DATE + timedelta(days=10)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == 0


def test_hold_last_value_filter_none() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=11)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series)


def test_hold_last_value_filter_some() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=3)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series)
    assert result["value"][1] == series["value"][0]
    assert result["ts"][1].as_py() == series["ts"][1].as_py()
    assert result["value"][2] == series["value"][0]
    assert result["ts"][2].as_py() == series["ts"][2].as_py()
    assert result["value"][3] == series["value"][0]
    assert result["ts"][3].as_py() == series["ts"][3].as_py()
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4].as_py() == series["ts"][4].as_py()


def test_hold_last_value_filter_some_with_quality() -> None:
    values = np.random.normal(size=10)
    quality = [0, 1] * 5
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=3)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series)
    assert result["value"][1] == series["value"][0]
    assert result["ts"][1].as_py() == series["ts"][1].as_py()
    assert result["quality"][1].as_py() == series["quality"][1].as_py()
    assert result["value"][2] == series["value"][0]
    assert result["ts"][2].as_py() == series["ts"][2].as_py()
    assert result["quality"][2].as_py() == series["quality"][2].as_py()
    assert result["value"][3] == series["value"][0]
    assert result["ts"][3].as_py() == series["ts"][3].as_py()
    assert result["quality"][3].as_py() == series["quality"][3].as_py()
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4].as_py() == series["ts"][4].as_py()
    assert result["quality"][4].as_py() == series["quality"][4].as_py()


def test_hold_last_value_filter_first() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE], [START_DATE + timedelta(hours=10)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series) - 1
    assert result["value"][0] == series["value"][1]
    assert result["ts"][0].as_py() == series["ts"][1].as_py()


def test_linear_interpolation_filter_all() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE], [START_DATE + timedelta(days=10)]
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == 0


def test_linear_interpolation_filter_none() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=11)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series)


def test_linear_interpolation_filter_some() -> None:
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=3)]
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series)
    assert result["value"][1].as_py() == _get_interpolation(
        series["value"][0].as_py(),
        series["value"][4].as_py(),
        series["ts"][0].as_py(),
        series["ts"][4].as_py(),
        series["ts"][1].as_py(),
    )
    assert result["ts"][1].as_py() == series["ts"][1].as_py()
    assert result["value"][2].as_py() == _get_interpolation(
        series["value"][0].as_py(),
        series["value"][4].as_py(),
        series["ts"][0].as_py(),
        series["ts"][4].as_py(),
        series["ts"][2].as_py(),
    )
    assert result["ts"][2].as_py() == series["ts"][2].as_py()
    assert result["value"][3].as_py() == _get_interpolation(
        series["value"][0].as_py(),
        series["value"][4].as_py(),
        series["ts"][0].as_py(),
        series["ts"][4].as_py(),
        series["ts"][3].as_py(),
    )
    assert result["ts"][3].as_py() == series["ts"][3].as_py()
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4].as_py() == series["ts"][4].as_py()


def test_linear_interpolation_filter_some_with_quality() -> None:
    values = np.random.normal(size=10)
    quality = [0, 1] * 5
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=3)]
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series)
    assert result["value"][1].as_py() == _get_interpolation(
        series["value"][0].as_py(),
        series["value"][4].as_py(),
        series["ts"][0].as_py(),
        series["ts"][4].as_py(),
        series["ts"][1].as_py(),
    )
    assert result["ts"][1].as_py() == series["ts"][1].as_py()
    assert result["quality"][1].as_py() == series["quality"][1].as_py()
    assert result["value"][2].as_py() == _get_interpolation(
        series["value"][0].as_py(),
        series["value"][4].as_py(),
        series["ts"][0].as_py(),
        series["ts"][4].as_py(),
        series["ts"][2].as_py(),
    )
    assert result["ts"][2].as_py() == series["ts"][2].as_py()
    assert result["quality"][2].as_py() == series["quality"][2].as_py()
    assert result["value"][3].as_py() == _get_interpolation(
        series["value"][0].as_py(),
        series["value"][4].as_py(),
        series["ts"][0].as_py(),
        series["ts"][4].as_py(),
        series["ts"][3].as_py(),
    )
    assert result["ts"][3].as_py() == series["ts"][3].as_py()
    assert result["quality"][3].as_py() == series["quality"][3].as_py()
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4].as_py() == series["ts"][4].as_py()
    assert result["quality"][4].as_py() == series["quality"][4].as_py()


def test_linear_interpolation_filter_first():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE], [START_DATE + timedelta(hours=10)]
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )

    assert len(result) == len(series) - 1
    assert result["value"][0] == series["value"][1]
    assert result["ts"][0].as_py() == series["ts"][1].as_py()


def test_linear_interpolation_filter_last_3():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=7)], [START_DATE + timedelta(days=10)]
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series) - 3
    assert result["value"][6] == series["value"][6]
    assert result["ts"][6].as_py() == series["ts"][6].as_py()


def test_linear_interpolation_filter_some_string_values():
    values = np.random.choice(np.array(list(string.ascii_lowercase + " ")), size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=3)]
    )

    result = filter_series(
        series, event_frames, AugmentationStrategy.LINEAR_INTERPOLATION
    )
    assert len(result) == len(series)
    assert result["value"][1] == series["value"][0]
    assert result["ts"][1].as_py() == series["ts"][1].as_py()
    assert result["value"][2] == series["value"][0]
    assert result["ts"][2].as_py() == series["ts"][2].as_py()
    assert result["value"][3] == series["value"][0]
    assert result["ts"][3].as_py() == series["ts"][3].as_py()
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4].as_py() == series["ts"][4].as_py()


def test_median_filter_second():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)],
        np.concatenate(([0] * 5, [1] * 5)),
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=1)],
        [START_DATE + timedelta(days=1) + timedelta(hours=10)],
    )

    result = filter_series(series, event_frames, AugmentationStrategy.MEAN)
    assert len(result) == len(series)
    assert result["value"][1].as_py() == 0.5


def test_median_filter_last_half():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)],
        np.concatenate(([0] * 5, [1] * 5)),
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=5)],
        [START_DATE + timedelta(days=5) + timedelta(days=5)],
    )

    result = filter_series(series, event_frames, AugmentationStrategy.MEAN)
    assert len(result) == len(series)
    assert result["value"][5].as_py() == 0.5
    assert result["value"][6].as_py() == 0.5


def test_median_filter_last_half_with_quality():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)],
        np.concatenate(([0] * 5, [1] * 5)),
        [0, 1] * 5,
    )
    event_frames = _create_event_frame_data(
        [START_DATE + timedelta(days=5)],
        [START_DATE + timedelta(days=5) + timedelta(days=5)],
    )

    result = filter_series(series, event_frames, AugmentationStrategy.MEAN)
    assert len(result) == len(series)
    assert result["value"][5].as_py() == 0.5
    assert result["value"][6].as_py() == 0.5
    assert result["quality"][1].as_py() == series["quality"][1].as_py()


def _get_interpolation(last_acceptable, next_value, last_time, next_time, current_time):
    numerator = current_time - last_time
    denominator = next_time - last_time
    time_interpolation = numerator / denominator
    return last_acceptable * (1 - time_interpolation) + next_value * time_interpolation



"""Timeseer Client provides convenient remote access to Timeseer.

The `Client` class creates a connection to Timeseer.

```
client = Client(api_key=(key_name, secret), host="timeseer.example.org")
```

`Client` supports basic time series operations:

```
client.list_sources()  # list all connected time series sources
client.search(...)  # search for time series
client.get_metadata(...)  # return metadata for a time series
client.get_data(...)  # return data for a time series
```

All concepts within Timeseer are represented as Resources.
The `Resources` class should be used to configure them.

```
resources = Resources(client)
resources.list(...)
resources.create(...)
resources.read(...)
resources.delete(...)
```

Several resources can be interacted with. They have their own class:

- `DataServices`
- `DataSets`
- `Flows`
- `Sources`

For example:

```
Flows(client).evaluate(...)
```

This package also contains several convenience functions to:

- Profile a data frame in Timeseer.
- Filter data based on event frames.

"""

from kukur import (
    DataType,
    Dictionary,
    InterpolationType,
    Metadata,
    SeriesSearch,
    SeriesSelector,
)

from timeseer_client.internal import (
    AugmentationStrategy,
    EventFrame,
    ProcessType,
    Statistic,
    TimeseerClientException,
    UnknownAugmentationStrategyException,
)
from timeseer_client.internal.client import Client
from timeseer_client.internal.data_services import DataServices, DataServiceSelector
from timeseer_client.internal.data_sets import DataSets
from timeseer_client.internal.data_substitutions import (
    apply_data_substitution,
    mask_event_frames,  # noqa: F401
)
from timeseer_client.internal.filters import filter_event_frames, filter_series
from timeseer_client.internal.flows import Flows
from timeseer_client.internal.profile import profile
from timeseer_client.internal.resources import Resources
from timeseer_client.internal.sources import Sources
from timeseer_client.metadata.fields import register_custom_fields

register_custom_fields(Metadata)


__all__ = [
    "AugmentationStrategy",
    "Client",
    "DataServices",
    "DataServiceSelector",
    "DataSets",
    "DataType",
    "Dictionary",
    "EventFrame",
    "Flows",
    "InterpolationType",
    "Metadata",
    "ProcessType",
    "Resources",
    "SeriesSearch",
    "SeriesSelector",
    "Sources",
    "Statistic",
    "TimeseerClientException",
    "UnknownAugmentationStrategyException",
    "filter_event_frames",
    "filter_series",
    "profile",
    "apply_data_substitution",
]



"""Python client for Timeseer Data Sets."""

from typing import Any, List, Tuple

import pyarrow as pa
from kukur import Metadata

from timeseer_client.internal import (
    JSONFlightClient,
    MissingTimezoneException,
    UnsupportedDataFormatException,
)

HAS_PANDAS = True
try:
    import pandas as pd

    from timeseer_client.internal import data_uploader_pandas

except Exception:
    HAS_PANDAS = False


class DataSets:
    """Data Sets are fixed collections of time series and their data in a specific time range.

    Args:
        client: the Timeseer Client
    """

    __client: JSONFlightClient

    def __init__(self, client: JSONFlightClient):
        self.__client = client

    def list(self) -> List[str]:
        """Return a list containing all the data set names."""
        return self.__client.do_action("data_sets/list", {})

    def remove_data(self, name: str):
        """Removes all data in a data set.

        Args:
            name: The name of the data set.
        """
        body = {"name": name}
        self.__client.do_action("data_sets/remove_data", body)

    def upload_data(
        self,
        many_series: List[Tuple[Metadata, Any]],
    ):
        """Upload time series data to a Data Set.

        The 'source' field in the SeriesSelector in Metadata determines the name of the data set.

        Data is provided as a pyarrow.Table or pandas DataFrame of two or three columns:
            The first column with name 'ts' contains Arrow timestamps.
            The second column with name 'value' contains the values as a number or string.
            The optional third column with name 'quality' contains a quality flag (0 is BAD, 1 is GOOD)

        Arguments:
            many_series: a list of tuple of metadata and data.
        """
        for metadata, table in many_series:
            if isinstance(table, pa.Table):
                data = _ensure_utc_timezone(table)
            elif HAS_PANDAS and isinstance(table, pd.DataFrame):
                data = data_uploader_pandas.to_pyarrow_table(table)
            else:
                raise UnsupportedDataFormatException()
            metadata_json = metadata.to_data()
            self.__client.do_put({"metadata": metadata_json}, data)


def _ensure_utc_timezone(table: pa.Table) -> pa.Table:
    if table.schema.field("ts").type.tz is None:
        raise MissingTimezoneException()

    return table.set_column(
        table.column_names.index("ts"),
        "ts",
        pa.compute.cast(table["ts"], pa.timestamp("us", "UTC")),
    )



"""Functions to handle data substitutions data."""

import pyarrow as pa

from timeseer_client.internal import DataSubstitutionData


def apply_data_substitution(
    original_data: pa.Table,
    data_substitution_data: DataSubstitutionData,
) -> pa.Table:
    """Apply a data substitution and return the resulting table."""
    has_quality = False
    substitution_data = data_substitution_data.data
    if "quality" in original_data.column_names:
        has_quality = True
        if has_quality and "quality" not in substitution_data.column_names:
            quality = pa.array([1] * len(substitution_data))
            substitution_data = substitution_data.append_column("quality", quality)
        if not has_quality and "quality" in substitution_data.column_names:
            substitution_data = substitution_data.drop_columns("quality")

    data = original_data.cast(substitution_data.schema)
    data_before = data.filter(
        pa.compute.less_equal(data["ts"], pa.scalar(data_substitution_data.start_date))
    )
    data_after = data.filter(
        pa.compute.greater_equal(data["ts"], pa.scalar(data_substitution_data.end_date))
    )

    return pa.concat_tables([data_before, substitution_data, data_after])


def mask_event_frames(data: pa.Table, event_frames: pa.Table) -> pa.Array:
    """Generate a boolean mask that is True for each data point in data were an event frame is active."""
    mask = pa.array([False] * len(data))

    for i in range(0, len(event_frames)):
        start_date = event_frames["start_date"][i]
        end_date = event_frames["end_date"][i]

        start_mask = pa.compute.greater_equal(data["ts"], start_date)
        end_mask = pa.compute.less(data["ts"], end_date)

        frame_mask = pa.compute.and_(start_mask, end_mask)
        mask = pa.compute.or_(mask, frame_mask)

    return mask



"""Python client for Timeseer Data Services."""

import time
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import pyarrow as pa
from kukur import Metadata, SeriesSelector

from timeseer_client.internal import (
    DataSubstitution,
    DataSubstitutionData,
    EventFrame,
    JSONFlightClient,
    Statistic,
    TimeoutException,
    TimeseerClientException,
)


class DataServiceEvaluationFailedException(TimeseerClientException):
    """Thrown when a data service evaluation fails."""

    def __init__(self, name: str, view_name: str):
        super().__init__(f'evaluation for data service "{name}"/"{view_name}" failed')


class DataServiceEvaluationsFailedException(TimeseerClientException):
    """Thrown when a data service evaluation fails."""

    def __init__(self, name: str):
        super().__init__(f'evaluations for data service "{name}" failed')


@dataclass
class DataServiceSelector:
    """Points to a series set inside a Data Service."""

    name: str
    series_set_name: str

    @classmethod
    def from_data(cls, data: Dict[str, str]) -> "DataServiceSelector":
        """Create a DataServiceSelector from a JSON dictionary."""
        return cls(data["name"], data["view"])

    def to_data(self) -> Dict[str, str]:
        """Convert to a dictionary suitable for JSON conversion."""
        return {"name": self.name, "view": self.series_set_name}


class DataServices:
    """Data Services provide access to analysis results, time series data and statistics.

    Args:
        client: the Timeseer Client
    """

    __client: JSONFlightClient

    def __init__(self, client: JSONFlightClient):
        self.__client = client

    def list(self) -> List[str]:
        """Return a list containing all the Data Service names."""
        return self.__client.do_action("data_services/list", {})

    def get(self, data_service_name: str) -> List[DataServiceSelector]:
        """Return a list containing a selector for each series set in a data service."""
        return [
            DataServiceSelector.from_data(data)
            for data in self.__client.do_action(
                "data_services/get", {"dataServiceName": data_service_name}
            )
        ]

    def wait_for_all_evaluations(
        self, data_service_name: str, *, timeout_seconds: float = 60.0
    ) -> None:
        """Wait until all the evaluations of a Data Service are complete.

        Args:
            data_service_name: The name of the Data Service to wait for.
            timeout_seconds: The number of seconds to wait until the evaluation is complete. (optional, keyword only)

        Raises:
            DataServiceEvaluationsFailedException when a failure is reported.
        """
        body = {"dataServiceName": data_service_name}
        while timeout_seconds > 0:
            job_states = self.__client.do_action(
                "data_services/get_all_view_states", body
            )
            if all(
                job_state["completed"] == job_state["total"] for job_state in job_states
            ):
                if any(job_state["failed"] > 0 for job_state in job_states):
                    raise DataServiceEvaluationsFailedException(data_service_name)
                return
            time.sleep(0.2)
            timeout_seconds = timeout_seconds - 0.2

        raise TimeoutException()

    def wait_for_evaluation(
        self, data_service: DataServiceSelector, *, timeout_seconds: int = 60
    ) -> None:
        """Wait until the evaluation of a Data Service is complete.

        Args:
            data_service: The series set in a Data Service to wait for.
            timeout_seconds: The number of seconds to wait until the evaluation is complete. (optional, keyword only)

        Raises:
            DataServiceEvaluationFailedException when a failure is reported.
        """
        body = {"dataService": data_service.to_data()}
        while timeout_seconds > 0:
            job_state = self.__client.do_action("data_services/get_view_state", body)
            if job_state["completed"] == job_state["total"]:
                if job_state["failed"] > 0:
                    raise DataServiceEvaluationFailedException(
                        data_service.name, data_service.series_set_name
                    )
                return
            time.sleep(1)
            timeout_seconds = timeout_seconds - 1

        raise TimeoutException()

    def get_data(
        self,
        data_service: DataServiceSelector,
        selector: SeriesSelector,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> pa.Table:
        """Get data from a given Data Service.

        Falls back to the data source when no data is present.

        Args:
            data_service: The series set in the Data Service to look for data in.
            selector: Return data for the time series selected by this selector.
            start_date: the start date of the data
                Defaults to start date of the data service.
            end_date: the end date of the data
                Defaults to end date of the data service.

        Returns::
            A pyarrow Table with three columns: 'ts', 'value' and 'quality'.
        """
        body = {
            "query": "data_services/get_data",
            "dataService": data_service.to_data(),
            "selector": selector.to_data(),
            "startDate": start_date.isoformat() if start_date is not None else None,
            "endDate": end_date.isoformat() if end_date is not None else None,
        }
        return self.__client.do_get(body)

    def list_data_substitutions(  # noqa: PLR0913
        self,
        data_service: DataServiceSelector,
        selector: SeriesSelector,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        last_updated_date: Optional[datetime] = None,
    ) -> List[DataSubstitution]:
        """List data substitutions for a Series in a data service.

        Args:
            data_service: The series set in the Data Service to look for data in.
            selector: Return data for the time series selected by this selector.
            start_date: The start date of the period of interest for the substitutions.
            end_date: The end date of the period of interest for the substitutions.
            last_updated_date: Only list substitutions updated on or after the specified date.

        Returns::
            A list of `DataSubstitution`s
        """
        body = {
            "dataService": data_service.to_data(),
            "selector": selector.to_data(),
            "startDate": start_date.isoformat() if start_date is not None else None,
            "endDate": end_date.isoformat() if end_date is not None else None,
            "lastUpdatedDate": (
                last_updated_date.isoformat() if last_updated_date is not None else None
            ),
        }
        return [
            DataSubstitution.from_data(substitution_data)
            for substitution_data in self.__client.do_action(
                "data_services/list_data_substitutions", body
            )
        ]

    def get_substitution_data(
        self, data_service: DataServiceSelector, data_substitution: DataSubstitution
    ) -> DataSubstitutionData:
        """Get the stored data for a given data substitution.

        Args:
            data_service: The series set in the Data Service to look for data in.
            data_substitution: The data substitution to query the data for.

        Returns::
            A pyarrow Table with three columns: 'ts', 'value' and 'quality'.
        """
        body = {
            "query": "data_services/get_substitution_data",
            "dataService": data_service.to_data(),
            "dataSubstitutionId": data_substitution.db_id,
        }
        return DataSubstitutionData(
            data_substitution.start_date,
            data_substitution.end_date,
            self.__client.do_get(body),
        )

    def get_kpi_scores(self, data_service: DataServiceSelector) -> Dict[str, int]:
        """Get the kpi scores of a Data Service.

        Args:
            data_service: The series set in the Data Service to return scores for.

        Returns::
            The score per KPI as a percentage.
        """
        body = {
            "dataService": data_service.to_data(),
        }
        return self.__client.do_action("data_services/get_kpi_scores", body)

    def convert_to_event_frame(self, table: pa.Table) -> List[EventFrame]:
        """Convert a PyArrow table to an EventFrame object."""
        return [EventFrame.from_row(data) for data in table.to_pylist()]

    def get_event_frames(  # noqa: PLR0913
        self,
        data_service: DataServiceSelector,
        selector: Optional[SeriesSelector] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        frame_type: Optional[Union[str, List[str]]] = None,
    ) -> pa.Table:
        """Get all event frames matching the given criteria.

        Args:
            data_service: The series set in the Data Service to return event frames for.
            selector: the time series to which the event frames are linked.
            start_date: the start date of the range to find overlapping event frames in.
                Defaults to start date of the data service.
            end_date: the end date of the range to find overlapping event frames in.
                Defaults to end date of the data service.
            frame_type: the type or types of event frames to search for. Finds all types when empty.

        Returns::
            A pyarrow Table with 6 columns.
            The first column ('start_date') contains the start date.
            The second column ('end_date') contains the end date.
            The third column ('type') contains the type of the returned event frame as a string.
            The fourth column ('explanation') can contain the explanation for an event frame as a string.
            The fifth column ('status') can contain the status of an event frame as a string.
            Columns 6 contains possible multiple references for the event frame.
            the seventh column contains the properties of the event frame.
            The eight column contains the uuid of the event frame.
        """
        query: Dict[str, Any] = {
            "query": "data_services/get_event_frames",
            "dataService": data_service.to_data(),
            "selector": selector.to_data() if selector is not None else None,
            "startDate": start_date.isoformat() if start_date is not None else None,
            "endDate": end_date.isoformat() if end_date is not None else None,
        }

        if frame_type is not None:
            query["type"] = frame_type

        return self.__client.do_get(query)

    def update_event_frame(self, event_frame: EventFrame):
        """Update the properties on an event frame.

        Args:
            event_frame: The event frame to be updated.
        """
        data = {
            "eventFrame": event_frame.to_data(),
        }
        return self.__client.do_action("data_services/update_event_frame", data)

    def list_series(self, data_service: DataServiceSelector) -> List[SeriesSelector]:
        """Return all series in a Data Service Series Set.

        Args:
            data_service: The series set in the Data Service to list series for.

        Returns::
            A list of `SeriesSelector`s.
        """
        query = {"dataService": data_service.to_data()}
        return [
            SeriesSelector.from_data(data)
            for data in self.__client.do_action("data_services/list_series", query)
        ]

    def get_statistics(
        self,
        data_service: DataServiceSelector,
        selector: Optional[SeriesSelector] = None,
    ) -> List[Statistic]:
        """Return statistics stored in a Data Service.

        Returns multivariate and calculated statistics for a series set if no SeriesSelector is provided.

        Args:
            data_service: The series set in the Data Service to return statistics for.
            selector: The time series to return statistics for. (optional, keyword-only)

        Returns::
            A list of `Statistic`s.
        """
        query: Dict[str, Any] = {"dataService": data_service.to_data()}

        if selector is not None:
            query["selector"] = selector.to_data()

        return [
            Statistic.from_data(data)
            for data in self.__client.do_action("data_services/get_statistics", query)
        ]

    def get_calculated_metadata(
        self, data_service: DataServiceSelector, selector: SeriesSelector
    ) -> Metadata:
        """Return calculated metadata for a series in a Data Service.

        Args:
            data_service: The series set in a Data Service where the calculated metadata is kept.
            selector: The time series to return calculated metadata for.

        Returns::
            A `Metadata` object with the calculated metadata that's available.
        """
        query = {
            "dataService": data_service.to_data(),
            "selector": selector.to_data(),
        }
        return Metadata.from_data(
            self.__client.do_action("data_services/get_calculated_metadata", query),
            selector,
        )



"""Implementation details for the Timeseer Client.

Only use classes and functions defined in timeseer_client.
"""

import json
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Any, Optional, Protocol, Union

import pyarrow as pa


class AugmentationStrategy(Enum):
    """AugmentationStrategy dictates what happens to filtered data points while augmenting based on event frames."""

    REMOVE = "remove values"
    HOLD_LAST = "hold last value"
    LINEAR_INTERPOLATION = "linear interpolation"
    KNN_IMPUTATION = "knn imputation"
    MEAN = "mean"


class TimeseerClientException(Exception):
    """Base class for Timeseer client exceptions.

    Use this to catch any exception that originates in the client.
    """


class AugmentationException(TimeseerClientException):
    """Exception raised when augmentation strategy fails."""


class UnknownAugmentationStrategyException(TimeseerClientException):
    """Raised when the augmentation strategy is not known."""


class MissingModuleException(TimeseerClientException):
    """Raised when a required Python module is not available."""

    def __init__(self, module_name: str):
        TimeseerClientException.__init__(
            self,
            f'missing Python package: "{module_name}"',
        )


class ServerReturnedException(TimeseerClientException):
    """Raised when the server returns an error in the response body."""

    def __init__(self, error: str):
        TimeseerClientException.__init__(
            self,
            f'Exception returned from the server: "{error}"',
        )


class MissingTimezoneException(TimeseerClientException):
    """Raised when a specified timeout is exceeded."""


class UnsupportedDataFormatException(TimeseerClientException):
    """Raised when the data to upload is not in a supported format."""


class TimeoutException(TimeseerClientException):
    """Raised when a specified timeout is exceeded."""


class ProcessType(Enum):
    """ProcessType represents the process type of a time series."""

    CONTINUOUS = "CONTINUOUS"
    REGIME = "REGIME"
    BATCH = "BATCH"
    COUNTER = "COUNTER"


@dataclass
class Statistic:
    """Statistic represents a statistics that has been calculated.

    Statistics have a name and data type and contain a result that is specific per data type.
    """

    name: str
    data_type: str
    result: Any

    @classmethod
    def from_data(cls, data: dict[str, Any]):
        """Create a Statistic from a data dictionary."""
        name: str = data["name"]
        data_type: str = data["dataType"]
        if data_type == "datetime":
            result: Any = datetime.fromisoformat(data["result"])
        else:
            result = data["result"]
        return Statistic(name, data_type, result)


@dataclass
class EventFrame:
    """EventFrame represents the event frames that were found."""

    start_date: datetime
    end_date: datetime
    type: str
    explanation: Optional[str]
    status: Optional[str]
    references: Optional[list[Any]]
    properties: Optional[dict[str, Union[str, float]]]
    uuid: str

    @classmethod
    def from_row(cls, data: dict[str, Any]):
        """Create a Statistic from a data dictionary."""
        return EventFrame(
            data["start_date"],
            data["end_date"],
            data["type"],
            data.get("explanation"),
            data.get("status"),
            data.get("reference"),
            data.get("properties"),
            data["uuid"],
        )

    def to_data(self) -> dict[str, Any]:
        """Convert to a dictionary suitable for JSON conversion."""
        end_date = None
        if self.end_date is not None:
            end_date = self.end_date.isoformat()
        return {
            "startDate": self.start_date.isoformat(),
            "endDate": end_date,
            "type": self.type,
            "explanation": self.explanation,
            "status": self.status,
            "references": self.references,
            "properties": json.dumps(self.properties),
            "uuid": self.uuid,
        }


@dataclass
class DataSubstitutionData:
    """Represents a stored data substitution."""

    start_date: datetime
    end_date: datetime
    data: pa.Table


@dataclass
class DataSubstitution:
    """Represents a data substitution on series data."""

    db_id: int
    start_date: datetime
    end_date: datetime
    strategy: str
    value: float | None
    limit: str | None
    last_updated_date: datetime | None

    @classmethod
    def from_data(cls, data: dict[str, Any]) -> "DataSubstitution":
        """Creates a new instance from a dict."""
        last_updated_date = None
        if "lastUpdateDate" in data:
            last_updated_date = datetime.fromisoformat(data["lastUpdatedDate"])

        return cls(
            data["id"],
            datetime.fromisoformat(data["startDate"]),
            datetime.fromisoformat(data["endDate"]),
            data["strategy"],
            data.get("value"),
            data.get("limit"),
            last_updated_date,
        )


class JSONFlightClient(Protocol):
    """Arrow Flight Client that uses JSON for actions."""

    def do_action(self, name: str, data: Any) -> Any:
        """Perform an Arrow Flight action with the given data provided as JSON."""

    def do_get(self, data: Any) -> pa.Table:
        """Return an Arrow Tabel for the given JSON-encoded ticket."""

    def do_put(self, data: Any, table: pa.Table) -> Any:
        """Do an Arrow Flight PUT request to upload an Arrow table."""



"""Python client for Timeseer Sources."""

import time

from timeseer_client.internal import JSONFlightClient


class Sources:
    """Sources provide access to timeseries and metadata.

    Args:
        client: the Timeseer Client
    """

    __client: JSONFlightClient

    def __init__(self, client: JSONFlightClient):
        self.__client = client

    def update(self, source_name: str, *, block=True):
        """Update a source.

        Args:
            source_name: the name of the source to analyze
            block: block until the update completes (keyword-only, default True)
        """
        self.__client.do_action("sources/update", {"sourceName": source_name})

        if block:
            self.wait_for_update(source_name)

    def wait_for_update(self, source_name: str):
        """Wait until the metadata update of the given source completes.

        Args:
            source_name: the name of the source to analyze
        """
        while True:
            state = self.__client.do_action(
                "sources/get_update_state", {"sourceName": source_name}
            )
            if state["completed"] == state["total"]:
                break
            time.sleep(0.2)



"""Convenience functions to profile a time series."""

from typing import Dict, List, Tuple

from kukur import Metadata, SeriesSelector
from pyarrow import Table

from timeseer_client.internal import JSONFlightClient
from timeseer_client.internal.data_services import DataServices, DataServiceSelector
from timeseer_client.internal.data_sets import DataSets
from timeseer_client.internal.flows import Flows
from timeseer_client.internal.resources import Resources


def profile(
    client: JSONFlightClient, name: str, many_series: List[Tuple[Metadata, Table]]
) -> List[Dict[str, str]]:
    """Profile time series.

    This creates a Data Set, Flow and Data Service with the given name.
    Waits until the flow has been evaluated.

    Args:
        client: The client for Timeseer.
        name: The name given to the data set, flow and data service that will be created.
        many_series: A list of metadata-data pairs for the time series to be profiled.

    Returns:
        The resources that have been created.
    """
    data_sets = DataSets(client)
    data_sets.remove_data(name)
    data_sets.upload_data(
        [(_ensure_source_name(name, metadata), data) for metadata, data in many_series]
    )

    resources = Resources(client)
    data_set_resource = resources.read({"type": "data set", "name": name})
    profile_resources = [
        {
            "type": "data service",
            "name": name,
            "kpiSet": "Data quality fundamentals",
            "range": data_set_resource["range"],
        },
        {
            "name": name,
            "dataSet": name,
            "blocks": [
                {"name": "Univariate analysis", "type": "Analysis"},
                {
                    "name": "Contribute to data service",
                    "type": "data_service_contribute",
                    "dataServiceName": name,
                    "contributionBlockNames": ["Univariate analysis"],
                    "reset": True,
                },
            ],
        },
    ]
    resources.create(profile_resources)

    flows = Flows(client)
    flows.evaluate(name)

    data_services = DataServices(client)
    data_services.wait_for_evaluation(DataServiceSelector(name, name))

    return [
        {"type": resource.get("type", "flow"), "name": resource["name"]}
        for resource in reversed([data_set_resource] + profile_resources)
    ]


def _ensure_source_name(name: str, metadata: Metadata) -> Metadata:
    selector = SeriesSelector(name, metadata.series.tags, metadata.series.field)
    new_metadata = Metadata(selector)
    for k, v in metadata.iter_names():
        new_metadata.set_field_by_name(k, v)
    return new_metadata



"""Functions to handle data uploads using Pandas."""

import pandas as pd
import pyarrow as pa

from timeseer_client.internal import MissingTimezoneException


def to_pyarrow_table(data: pd.DataFrame) -> pa.Table:
    """Converts a pandas dataframe to pyarrow."""
    data["ts"] = _ensure_utc_timezone(data["ts"])
    return pa.Table.from_pandas(data, preserve_index=False)


def _ensure_utc_timezone(ts: pd.Series):
    if ts.dt.tz is None:
        raise MissingTimezoneException()
    return ts.dt.tz_convert("UTC")



"""Filter event frames in Arrow tables or merge Arrow-backed event frame lists."""

from datetime import datetime
from typing import List

import pyarrow as pa
import pyarrow.compute as pc

from timeseer_client.internal import AugmentationException, AugmentationStrategy


def filter_series(
    series: pa.Table,
    event_frames: pa.Table,
    augmentation_strategy: AugmentationStrategy,
) -> pa.Table:
    """Filter the even_frames out of the series based on the strategy."""
    if augmentation_strategy == AugmentationStrategy.KNN_IMPUTATION:
        raise AugmentationException()

    total = pa.array([True] * series.num_rows)

    for index, frame in enumerate(event_frames["start_date"]):
        on_or_after = pc.less(series["ts"], frame)
        before = pc.greater(series["ts"], event_frames["end_date"][index])
        out_event_frame = pc.or_(on_or_after, before)
        total = pc.and_(total, out_event_frame)

    if augmentation_strategy == AugmentationStrategy.HOLD_LAST:
        return _hold_last_value(series, total)

    if augmentation_strategy == AugmentationStrategy.LINEAR_INTERPOLATION:
        return _linear_interpolation(series, total)

    if augmentation_strategy == AugmentationStrategy.MEAN:
        return _mean_imputation(series, total)

    return series.filter(total)


def filter_event_frames(
    event_frames: pa.Table, start_date: datetime, end_date: datetime
) -> pa.Table:
    """Restrict time range to filter event frames."""
    if event_frames.num_rows == 0:
        return event_frames

    new_start_dates = []
    new_end_dates = []
    total: List[bool] = [True] * event_frames.num_rows
    for index, start_date_frame in enumerate(event_frames["start_date"]):
        end_date_frame = event_frames["end_date"][index]

        out_start = pc.greater_equal(start_date_frame, pa.scalar(end_date))
        out_end = pc.less_equal(end_date_frame, pa.scalar(start_date))
        new_start_date = start_date_frame.as_py()
        new_end_date = end_date_frame.as_py()
        if out_start.as_py() or out_end.as_py():
            total[index] = False
        if pc.less(start_date_frame, pa.scalar(start_date)).as_py():
            new_start_date = start_date
        if pc.greater(end_date_frame, pa.scalar(end_date)).as_py():
            new_end_date = end_date
        new_start_dates.append(new_start_date)
        new_end_dates.append(new_end_date)
    new_table = event_frames.set_column(
        0, "start_date", pa.array(new_start_dates, type=pa.timestamp("us", tz="UTC"))
    )
    new_table = new_table.set_column(
        1, "end_date", pa.array(new_end_dates, type=pa.timestamp("us", tz="UTC"))
    )
    return new_table.filter(total)


def _hold_last_value(series: pa.Table, out_of_event_frames: pa.Array) -> pa.Table:
    new_values = []
    new_ts = []
    new_quality = []
    last_acceptable = None
    for index, out_of_frame in enumerate(out_of_event_frames.to_pylist()):
        new_value = series["value"][index].as_py()

        if out_of_frame:
            last_acceptable = series["value"][index].as_py()
        else:
            if last_acceptable is None:
                continue
            new_value = last_acceptable

        new_values.append(new_value)
        new_ts.append(series["ts"][index].as_py())
        if "quality" in series.column_names:
            new_quality.append(series["quality"][index].as_py())
    if "quality" in series.column_names:
        return pa.Table.from_arrays(
            [pa.array(new_ts), pa.array(new_values), pa.array(new_quality)],
            ["ts", "value", "quality"],
        )
    return pa.Table.from_arrays(
        [pa.array(new_ts), pa.array(new_values)],
        ["ts", "value"],
    )


def _linear_interpolation(series: pa.Table, out_of_event_frames: pa.Array) -> pa.Table:
    new_values = []
    new_ts = []
    new_quality = []
    series_to_filter_count = 0
    last_acceptable = None
    for index, out_of_frame in enumerate(out_of_event_frames.to_pylist()):
        if out_of_frame:
            if series_to_filter_count != 0 and last_acceptable is not None:
                for element in range(series_to_filter_count):
                    numerator = (
                        series["ts"][index - (series_to_filter_count - element)].as_py()
                        - series["ts"][index - (series_to_filter_count + 1)].as_py()
                    )
                    denominator = (
                        series["ts"][index].as_py()
                        - series["ts"][index - (series_to_filter_count + 1)].as_py()
                    )
                    time_interpolation = numerator / denominator
                    new_value = (
                        last_acceptable * (1 - time_interpolation)
                        + series["value"][index].as_py() * time_interpolation
                    )
                    new_values.append(new_value)
                    new_ts.append(
                        series["ts"][index - (series_to_filter_count - element)].as_py()
                    )
                    if "quality" in series.column_names:
                        new_quality.append(
                            series["quality"][
                                index - (series_to_filter_count - element)
                            ].as_py()
                        )

            last_acceptable = series["value"][index].as_py()
            new_ts.append(series["ts"][index].as_py())
            new_values.append(last_acceptable)
            if "quality" in series.column_names:
                new_quality.append(series["quality"][index].as_py())
            series_to_filter_count = 0
        else:
            series_to_filter_count = series_to_filter_count + 1

    if "quality" in series.column_names:
        return pa.Table.from_arrays(
            [pa.array(new_ts), pa.array(new_values), pa.array(new_quality)],
            ["ts", "value", "quality"],
        )

    return pa.Table.from_arrays(
        [pa.array(new_ts), pa.array(new_values)], ["ts", "value"]
    )


def _mean_imputation(series: pa.Table, out_of_event_frames: pa.Array) -> pa.Table:
    median_value = pc.mean(series["value"])
    new_values = []
    new_ts = []
    new_quality = []
    last_acceptable = median_value.as_py()
    for index, out_of_frame in enumerate(out_of_event_frames.to_pylist()):
        new_value = series["value"][index].as_py()

        if out_of_frame:
            last_acceptable = median_value.as_py()
        else:
            if last_acceptable is None:
                continue
            new_value = last_acceptable

        new_values.append(new_value)
        new_ts.append(series["ts"][index].as_py())
        if "quality" in series.column_names:
            new_quality.append(series["quality"][index].as_py())

    if "quality" in series.column_names:
        return pa.Table.from_arrays(
            [pa.array(new_ts), pa.array(new_values), pa.array(new_quality)],
            ["ts", "value", "quality"],
        )
    return pa.Table.from_arrays(
        [pa.array(new_ts), pa.array(new_values)], ["ts", "value"]
    )



from datetime import datetime, timedelta

import pyarrow as pa

from timeseer_client import filter_event_frames


def _create_event_frame_data(start_dates, end_dates):
    series_type = pa.struct(
        [
            ("series_source", pa.string()),
            ("series_name", pa.string()),
        ]
    )
    references_type = pa.list_(series_type)
    types = []
    references = []
    types.append("type")
    references.append(None)
    return pa.Table.from_arrays(
        [
            pa.array(start_dates, type=pa.timestamp("us", tz="+00:00")),
            pa.array(end_dates, type=pa.timestamp("us", tz="+00:00")),
            pa.array(types),
            pa.array(references, type=references_type),
        ],
        ["start_date", "end_date", "type", "reference"],
    )


class TestFilterSeries:
    start_date = datetime.fromisoformat("2020-01-01T00:00:00.000+00:00")
    end_date = datetime.fromisoformat("2020-02-01T00:00:00.000+00:00")

    def test_filter_all(self):
        event_frames = _create_event_frame_data(
            [self.end_date], [self.end_date + timedelta(days=11)]
        )

        result = filter_event_frames(event_frames, self.start_date, self.end_date)
        assert len(result) == 0

    def test_filter_none(self):
        event_frames = _create_event_frame_data(
            [self.start_date + timedelta(days=11)], [self.end_date - timedelta(days=12)]
        )

        result = filter_event_frames(event_frames, self.start_date, self.end_date)
        assert len(result) == len(event_frames)
        assert result["start_date"][0].as_py() == self.start_date + timedelta(days=11)
        assert result["end_date"][0].as_py() == self.end_date - timedelta(days=12)

    def test_filter_change_start_date(self):
        event_frames = _create_event_frame_data(
            [self.start_date - timedelta(days=11)], [self.end_date - timedelta(days=12)]
        )

        result = filter_event_frames(event_frames, self.start_date, self.end_date)
        assert len(result) == len(event_frames)
        assert result["start_date"][0].as_py() == self.start_date
        assert result["end_date"][0].as_py() == self.end_date - timedelta(days=12)

    def test_filter_change_end_date(self):
        event_frames = _create_event_frame_data(
            [self.start_date + timedelta(days=11)], [self.end_date + timedelta(days=12)]
        )

        result = filter_event_frames(event_frames, self.start_date, self.end_date)
        assert len(result) == len(event_frames)
        assert result["start_date"][0].as_py() == self.start_date + timedelta(days=11)
        assert result["end_date"][0].as_py() == self.end_date



"""Test series filtering when providing Pandas DataFrames"""

from datetime import datetime, timedelta

import numpy as np
import pandas as pd
import pytest
from kukur import InterpolationType

from timeseer_client import (
    AugmentationStrategy,
    UnknownAugmentationStrategyException,
    filter_series,
)

START_DATE = datetime.fromisoformat("2020-01-01T00:00:00.000+00:00")


def _create_series_data(timestamps, values, quality=None) -> pd.DataFrame:
    if quality is not None:
        return pd.DataFrame(
            data={"ts": timestamps, "value": values, "quality": quality}
        )
    return pd.DataFrame(data={"ts": timestamps, "value": values})


def _create_event_frame_data(types, start_dates, end_dates) -> pd.DataFrame:
    return pd.DataFrame(
        data={"type": types, "start_date": start_dates, "end_date": end_dates}
    )


def test_filter_empty_series():
    series = _create_series_data([], [])
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE], [START_DATE + timedelta(days=10)]
    )
    result = filter_series(series, event_frames)
    assert len(result) == 0


def test_filter_empty_event_frames():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], np.random.normal(size=10)
    )
    event_frames = _create_event_frame_data([], [], [])
    result = filter_series(series, event_frames)
    assert len(result) == 10


def test_filter_invalid():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], np.random.normal(size=10)
    )
    event_frames = _create_event_frame_data([], [], [])

    with pytest.raises(UnknownAugmentationStrategyException):
        filter_series(series, event_frames, "knn imputation")


def test_filter_all():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE], [START_DATE + timedelta(days=10)]
    )
    result = filter_series(series, event_frames)
    assert len(result) == 0


def test_filter_none():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=11)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == len(series)


def test_filter_one():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == 1
    assert result["ts"][0] == START_DATE


def test_filter_one_with_quality():
    values = np.random.normal(size=10)
    quality = [0, 1] * 5
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == 1
    assert result["ts"][0] == START_DATE
    assert result["quality"][0] == series["quality"][0]


def test_filter_datetime_index():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    ).set_index("ts")
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=11)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames)
    assert len(result) == len(series)
    assert len(result.columns) == 1
    assert result.index[0] == START_DATE


def test_hold_last_value_filter_all():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE], [START_DATE + timedelta(days=10)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == 0


def test_hold_last_value_filter_none():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=11)], [START_DATE + timedelta(days=12)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series)


def test_hold_last_value_filter_some():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=3)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series)
    assert result["value"][1] == series["value"][0]
    assert result["ts"][1] == series["ts"][1]
    assert result["value"][2] == series["value"][0]
    assert result["ts"][2] == series["ts"][2]
    assert result["value"][3] == series["value"][0]
    assert result["ts"][3] == series["ts"][3]
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4] == series["ts"][4]


def test_hold_last_value_filter_some_with_quality():
    values = np.random.normal(size=10)
    quality = [0, 1] * 5
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=1)], [START_DATE + timedelta(days=3)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series)
    assert result["value"][1] == series["value"][0]
    assert result["ts"][1] == series["ts"][1]
    assert result["quality"][1] == series["quality"][1]
    assert result["value"][2] == series["value"][0]
    assert result["ts"][2] == series["ts"][2]
    assert result["quality"][2] == series["quality"][2]
    assert result["value"][3] == series["value"][0]
    assert result["ts"][3] == series["ts"][3]
    assert result["quality"][3] == series["quality"][3]
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4] == series["ts"][4]
    assert result["quality"][4] == series["quality"][4]


def test_hold_last_value_filter_first():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE], [START_DATE + timedelta(hours=10)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.HOLD_LAST)
    assert len(result) == len(series) - 1
    assert result["value"][0] == series["value"][1]
    assert result["ts"][0] == series["ts"][1]


def test_linear_interpolation_filter_all():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE, START_DATE + timedelta(days=10)
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == 0


def test_linear_interpolation_filter_none():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE + timedelta(days=11), START_DATE + timedelta(days=12)
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series)


def test_linear_interpolation_filter_some():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE + timedelta(days=1), START_DATE + timedelta(days=3)
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series)
    assert result["value"][1] == _get_interpolation(
        series["value"][0],
        series["value"][4],
        series["ts"][0],
        series["ts"][4],
        series["ts"][1],
    )
    assert result["ts"][1] == series["ts"][1]
    assert result["value"][2] == _get_interpolation(
        series["value"][0],
        series["value"][4],
        series["ts"][0],
        series["ts"][4],
        series["ts"][2],
    )
    assert result["ts"][2] == series["ts"][2]
    assert result["value"][3] == _get_interpolation(
        series["value"][0],
        series["value"][4],
        series["ts"][0],
        series["ts"][4],
        series["ts"][3],
    )
    assert result["ts"][3] == series["ts"][3]
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4] == series["ts"][4]


def test_linear_interpolation_filter_some_qith_quality():
    values = np.random.normal(size=10)
    quality = [0, 1] * 5
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE + timedelta(days=1), START_DATE + timedelta(days=3)
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series)
    assert result["value"][1] == _get_interpolation(
        series["value"][0],
        series["value"][4],
        series["ts"][0],
        series["ts"][4],
        series["ts"][1],
    )
    assert result["ts"][1] == series["ts"][1]
    assert result["quality"][1] == series["quality"][1]
    assert result["value"][2] == _get_interpolation(
        series["value"][0],
        series["value"][4],
        series["ts"][0],
        series["ts"][4],
        series["ts"][2],
    )
    assert result["ts"][2] == series["ts"][2]
    assert result["quality"][2] == series["quality"][2]
    assert result["value"][3] == _get_interpolation(
        series["value"][0],
        series["value"][4],
        series["ts"][0],
        series["ts"][4],
        series["ts"][3],
    )
    assert result["ts"][3] == series["ts"][3]
    assert result["quality"][3] == series["quality"][3]
    assert result["value"][4] == series["value"][4]
    assert result["ts"][4] == series["ts"][4]
    assert result["quality"][4] == series["quality"][4]


def test_linear_interpolation_filter_first():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE, START_DATE + timedelta(hours=10)
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series) - 1
    assert result["value"][0] == series["value"][1]
    assert result["ts"][0] == series["ts"][1]


def test_linear_interpolation_filter_last_3():
    values = np.random.normal(size=10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE + timedelta(days=7), START_DATE + timedelta(days=10)
    )

    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.LINEAR_INTERPOLATION,
        InterpolationType.LINEAR,
    )
    assert len(result) == len(series) - 3
    assert result["value"][6] == series["value"][6]
    assert result["ts"][6] == series["ts"][6]


def test_knn_imputation_at_beginning():
    values = np.arange(10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    context = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE, START_DATE + timedelta(hours=10)
    )
    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.KNN_IMPUTATION,
        InterpolationType.LINEAR,
        [context],
    )
    assert len(result) == len(series)
    assert result["value"][0] == 3
    assert result["ts"][6] == series["ts"][6]


def test_knn_imputation_at_beginning_with_quality():
    values = np.arange(10)
    quality = [0, 1] * 5
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    context = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values, quality
    )
    event_frames = _create_event_frame_data(
        ["random"], START_DATE, START_DATE + timedelta(hours=10)
    )
    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.KNN_IMPUTATION,
        InterpolationType.LINEAR,
        [context],
    )
    assert len(result) == len(series)
    assert result["value"][0] == 3
    assert result["ts"][6] == series["ts"][6]
    assert result["quality"][6] == series["quality"][6]


def test_knn_imputation_at_beginning_multiple_context():
    values = np.arange(10)
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)], values
    )
    context = [
        _create_series_data(
            [START_DATE + timedelta(days=i) for i in range(10)], values
        ),
        _create_series_data(
            [START_DATE + timedelta(days=i) for i in range(10)], values + 10
        ),
    ]

    event_frames = _create_event_frame_data(
        ["random"], START_DATE, START_DATE + timedelta(hours=10)
    )
    result = filter_series(
        series,
        event_frames,
        AugmentationStrategy.KNN_IMPUTATION,
        InterpolationType.LINEAR,
        context,
    )
    assert len(result) == len(series)
    assert result["value"][0] == 3
    assert result["ts"][6] == series["ts"][6]


def test_median_filter_second():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)],
        np.concatenate(([0] * 5, [1] * 5)),
    )
    event_frames = _create_event_frame_data(
        ["random"],
        [START_DATE + timedelta(days=1)],
        [START_DATE + timedelta(days=1) + timedelta(hours=10)],
    )

    result = filter_series(series, event_frames, AugmentationStrategy.MEAN)
    assert len(result) == len(series)
    assert result["value"][1] == 0.5


def test_median_filter_last_half():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)],
        np.concatenate(([0] * 5, [1] * 5)),
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=5)], [START_DATE + timedelta(days=10)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.MEAN)
    assert len(result) == len(series)
    assert result["value"][0] == 0
    assert result["value"][5] == 0.5
    assert result["value"][6] == 0.5


def test_median_filter_last_half_with_quality():
    series = _create_series_data(
        [START_DATE + timedelta(days=i) for i in range(10)],
        np.concatenate(([0] * 5, [1] * 5)),
        [0, 1] * 5,
    )
    event_frames = _create_event_frame_data(
        ["random"], [START_DATE + timedelta(days=5)], [START_DATE + timedelta(days=10)]
    )

    result = filter_series(series, event_frames, AugmentationStrategy.MEAN)
    assert len(result) == len(series)
    assert result["value"][0] == 0
    assert result["quality"][0] == series["quality"][0]
    assert result["value"][5] == 0.5
    assert result["quality"][5] == series["quality"][5]
    assert result["value"][6] == 0.5
    assert result["quality"][6] == series["quality"][6]


def _get_interpolation(last_acceptable, next_value, last_time, next_time, current_time):
    numerator = current_time - last_time
    denominator = next_time - last_time
    time_interpolation = numerator / denominator
    return last_acceptable * (1 - time_interpolation) + next_value * time_interpolation



"""Filter event frames in Pandas DataFrames tables or merge DataFrame-backed event frame lists."""

from datetime import datetime
from typing import List

import numpy as np
import pandas as pd

try:
    from sklearn.impute import KNNImputer

    HAS_SKLEARN = True
except Exception:
    HAS_SKLEARN = False

from timeseer_client.internal import AugmentationException, AugmentationStrategy


def filter_series(
    series: pd.DataFrame,
    event_frames: pd.DataFrame,
    augmentation_strategy: AugmentationStrategy,
    context: List[pd.DataFrame],
) -> pd.DataFrame:
    """Filter the even_frames out of the series with the given strategy."""
    has_datetime_index = False
    if len(series.columns) == 1:
        has_datetime_index = True
        series = series.reset_index()

    if augmentation_strategy == AugmentationStrategy.KNN_IMPUTATION:
        if len(context) == 0:
            raise AugmentationException("KNN imputation requires context.")
        if not HAS_SKLEARN:
            raise AugmentationException(
                "sklearn is required to perform KNN imputation."
            )

    out_event_frame = [True] * len(series)
    for index, start_date in enumerate(event_frames["start_date"]):
        out_event_frame = out_event_frame & (
            (series["ts"] < start_date)
            | (series["ts"] > event_frames["end_date"][index])
        )

    if augmentation_strategy == AugmentationStrategy.HOLD_LAST:
        result = _hold_last_value(series, out_event_frame)

    if augmentation_strategy == AugmentationStrategy.LINEAR_INTERPOLATION:
        result = _linear_interpolation(series, out_event_frame)

    if augmentation_strategy == AugmentationStrategy.KNN_IMPUTATION:
        result = _knn_imputation(series, out_event_frame, context)

    if augmentation_strategy == AugmentationStrategy.MEAN:
        result = _mean_imputation(series, out_event_frame)

    if augmentation_strategy == AugmentationStrategy.REMOVE:
        result = series.loc[out_event_frame]

    if has_datetime_index:
        return result.set_index("ts", drop=True)
    return result


def filter_event_frames(
    event_frames: pd.DataFrame, start_date: datetime, end_date: datetime
) -> pd.DataFrame:
    """Restrict time range to filter event frames."""
    if len(event_frames) == 0:
        return event_frames

    filtered = event_frames.loc[
        (event_frames["start_date"] > pd.to_datetime(end_date))
        | (event_frames["end_date"] > pd.to_datetime(start_date))
    ].copy()

    filtered.loc[
        filtered["start_date"] < pd.to_datetime(start_date), "start_date"
    ] = pd.to_datetime(start_date)
    filtered.loc[
        filtered["end_date"] > pd.to_datetime(end_date), "end_date"
    ] = pd.to_datetime(end_date)
    return filtered


def _hold_last_value(
    series: pd.DataFrame, out_of_event_frames: List[bool]
) -> pd.DataFrame:
    last_acceptable = None
    count = 0
    for index, out_of_frame in enumerate(out_of_event_frames):
        if out_of_frame:
            last_acceptable = series["value"][index]
        else:
            if last_acceptable is None:
                count = count + 1
            series.loc[index, "value"] = last_acceptable
    series = series.loc[count:]
    return series.reset_index(drop=True)


def _linear_interpolation(
    series: pd.DataFrame, out_of_event_frames: List[bool]
) -> pd.DataFrame:
    series_to_filter_count = 0
    last_acceptable = None
    remove_first_elements = 0
    for index, out_of_frame in enumerate(out_of_event_frames):
        if out_of_frame:
            if series_to_filter_count != 0 and last_acceptable is not None:
                for element in range(series_to_filter_count):
                    numerator = (
                        series["ts"][index - (series_to_filter_count - element)]
                        - series["ts"][index - (series_to_filter_count + 1)]
                    )
                    denominator = (
                        series["ts"][index]
                        - series["ts"][index - (series_to_filter_count + 1)]
                    )
                    time_interpolation = numerator / denominator
                    new_value = (
                        last_acceptable * (1 - time_interpolation)
                        + series["value"][index] * time_interpolation
                    )
                    series.loc[
                        index - (series_to_filter_count - element), "value"
                    ] = new_value
            if last_acceptable is None:
                remove_first_elements = series_to_filter_count

            last_acceptable = series["value"][index]
            series_to_filter_count = 0
        else:
            series_to_filter_count = series_to_filter_count + 1

    series = series.loc[
        remove_first_elements : (len(out_of_event_frames) - series_to_filter_count - 1)
    ]
    return series.reset_index(drop=True)


def _knn_imputation(
    original_series: pd.DataFrame,
    out_of_event_frames: List[bool],
    context: List[pd.DataFrame],
):
    original_series.loc[~np.array(out_of_event_frames), ("value")] = np.nan
    series = original_series.set_index("ts", drop=True)[["value"]]
    context = [serie.set_index("ts", drop=True)[["value"]] for serie in context]

    concatenated_df = pd.concat(
        [series, *context], axis=1, sort=False, ignore_index=True
    )
    imputer = KNNImputer(n_neighbors=5)
    imputed_df = pd.DataFrame(
        data=imputer.fit_transform(concatenated_df), index=concatenated_df.index
    )
    imputed_df = imputed_df.loc[series.index, 0].reset_index()
    if "quality" in original_series.columns:
        imputed_df["quality"] = original_series["quality"]
        imputed_df.columns = ["ts", "value", "quality"]
    else:
        imputed_df.columns = ["ts", "value"]
    return imputed_df


def _mean_imputation(
    series: pd.DataFrame, out_of_event_frames: List[bool]
) -> pd.DataFrame:
    mean = np.mean(series["value"])
    series.loc[~np.array(out_of_event_frames), ("value")] = np.nan
    series = series.set_index("ts", drop=True)
    series = series.fillna(mean)
    series.reset_index()
    return series



"""Python client for Timeseer Flows."""

import time
from datetime import datetime
from typing import Dict, List, Optional

from kukur import SeriesSelector

from timeseer_client.internal import JSONFlightClient, TimeseerClientException


class FlowEvaluationFailedException(TimeseerClientException):
    """Thrown when a flow evaluation fails."""

    def __init__(self, flow_name: str | None):
        if flow_name is None:
            super().__init__("flow evaluation failed.")
        else:
            super().__init__(f'evaluation for flow "{flow_name}" failed.')


class Flows:
    """Flows run modules to process time series data.

    Args:
        client: the Timeseer Client
    """

    __client: JSONFlightClient

    def __init__(self, client: JSONFlightClient):
        self.__client = client

    def list(self) -> List[str]:
        """Return a list containing all the flow names."""
        return self.__client.do_action("flows/list", {})

    def evaluate(self, flow_name: str, *, block=True):
        """Evaluate a flow.

        Args:
            flow_name: the name of the flow to evaluate
            block: block until the evaluation completes (keyword-only, default True)

        Raises:
            FlowEvaluationFailedException when a failure is reported.
        """
        evaluate_flow(self.__client, flow_name, block=block)

    def duplicate(  # noqa: PLR0913
        self,
        existing_flow_name: str,
        new_flow_name: str,
        *,
        series_set_names: Optional[List[str]] = None,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        data_set_name: Optional[str] = None,
    ):
        """Duplicate an existing flow.

        Args:
            existing_flow_name: the name of the flow to duplicate.
            new_flow_name: the name of the duplicated flow.
            series_set_names: the names of the existing series sets to be used in the flow. (Optional).
            start_date: the start date of the flow (Optional).
            end_date: the end date of the flow (Optional).
            data_set_name: the name of the data set to use for the flow (Optional).
        """
        body: Dict = {
            "existingFlowName": existing_flow_name,
            "newFlowName": new_flow_name,
            "seriesSetNames": series_set_names,
            "dataSetName": data_set_name,
        }
        if start_date is not None:
            body["startDate"] = start_date.isoformat()
        if end_date is not None:
            body["endDate"] = end_date.isoformat()
        self.__client.do_action("flows/duplicate", body)


def evaluate_flow(
    client: JSONFlightClient,
    flow_name: str,
    *,
    limitations: Optional[List[SeriesSelector]] = None,
    block=True,
):
    """Trigger an evaluation of the flow with the given name."""
    flow: Dict = {"flowName": flow_name}
    if limitations is not None:
        flow["limitations"] = [selector.to_data() for selector in limitations]
    response = client.do_action("flows/evaluate", flow)
    if block:
        wait_for_flow_evaluation(client, response, flow_name)


def wait_for_flow_evaluation(
    client: JSONFlightClient, response: Dict, flow_name: str | None = None
):
    """Repeatedly query for the flow evaluation state of the given flow evaluation group."""
    while True:
        state = client.do_action("flows/get_evaluation_state", response)
        if (
            state["completed"] == state["total"]
            and state["blockCompleted"] == state["blockTotal"]
        ):
            if (
                state["failed"] > 0
                or state["blockFailed"] > 0
                or state["currentBlockFailed"] > 0
            ):
                raise FlowEvaluationFailedException(flow_name)
            break
        time.sleep(0.2)
